@@ -2,7 +2,7 @@ package Config::IniFiles;
 
 use vars qw($VERSION);
 
-$VERSION = "2.52";
+$VERSION = '2.68';
 
 require 5.004;
 use strict;
@@ -20,7 +20,7 @@ Config::IniFiles - A module for reading .ini-style configuration files.
 =head1 SYNOPSIS
 
   use Config::IniFiles;
-  my $cfg = new Config::IniFiles( -file => "/path/configfile.ini" );
+  my $cfg = Config::IniFiles->new( -file => "/path/configfile.ini" );
   print "The value is " . $cfg->val( 'Section', 'Parameter' ) . "."
   	if $cfg->val( 'Section', 'Parameter' );
 
@@ -119,26 +119,31 @@ any of the following things:
 
   1) the pathname of a file
 
-    $cfg = new Config::IniFiles -file => "/path/to/config_file.ini";
+    $cfg = Config::IniFiles->new( -file => "/path/to/config_file.ini" );
 
   2) a simple filehandle
 
-    $cfg = new Config::IniFiles -file => STDIN;
+    $cfg = Config::IniFiles->new( -file => STDIN );
 
   3) a filehandle glob
 
     open( CONFIG, "/path/to/config_file.ini" );
-    $cfg = new Config::IniFiles -file => *CONFIG;
+    $cfg = Config::IniFiles->new( -file => *CONFIG );
 
   4) a reference to a glob
 
     open( CONFIG, "/path/to/config_file.ini" );
-    $cfg = new Config::IniFiles -file => \*CONFIG;
+    $cfg = Config::IniFiles->new( -file => \*CONFIG );
 
   5) an IO::File object
 
-    $io = new IO::File( "/path/to/config_file.ini" );
-    $cfg = new Config::IniFiles -file => $io;
+    $io = IO::File->new( "/path/to/config_file.ini" );
+    $cfg = Config::IniFiles->new( -file => $io );
+
+  or
+
+    open my $fh, '<', "/path/to/config_file.ini" or die $!;
+    $cfg = Config::IniFiles->new( -file => $fh );
 
   6) A reference to a scalar (requires newer versions of IO::Scalar)
 
@@ -148,7 +153,7 @@ any of the following things:
     Setting=Another value
     EOT
     
-    $cfg = new Config::IniFiles -file => $ini_file_contents;
+    $cfg = Config::IniFiles->new( -file => \$ini_file_contents );
 
 
 If this option is not specified, (i.e. you are creating a config file from scratch) 
@@ -173,7 +178,7 @@ in the "joe" section, there is none.
 
 If you create your Config::IniFiles object with a default section of "all" like this:
 
-   $cfg = new Config::IniFiles -file => "file.ini", -default => "all";
+   $cfg = Config::IniFiles->new( -file => "file.ini", -default => "all" );
    
 Then requsting a value for a "permissions" in the [joe] section will 
 check for a value from [all] before returning undef.
@@ -181,6 +186,29 @@ check for a value from [all] before returning undef.
    $permissions = $cfg->val( "joe", "permissions");   // returns "Nothing"
 
 
+=item I<-fallback> section
+
+Specifies a section to be used for parameters outside a section. Default is none.
+Without -fallback specified (which is the default), reading a configuration file
+which has a parameter outside a section will fail. With this set to, say,
+"GENERAL", this configuration:
+
+   wrong=wronger
+
+   [joe]
+   name=Joseph
+
+will be assumed as:
+
+   [GENERAL]
+   wrong=wronger
+
+   [joe]
+   name=Joseph
+
+Note that Config::IniFiles will also omit the fallback section header when
+outputing such configuration.
+
 =item I<-nocase> 0|1
 
 Set -nocase => 1 to handle the config file in a case-insensitive
@@ -280,6 +308,40 @@ or daemon application. The application is still responsible for determining
 when the object is to be reloaded.
 
 
+=item I<-nomultiline> 0|1
+
+Set -nomultiline => 1 to output multi-valued parameter as:
+
+ param=value1
+ param=value2
+
+instead of the default:
+
+ param=<<EOT
+ value1
+ value2
+ EOT
+
+As the later might not be compatible with all applications.
+
+=item I<-handle_trailing_comment> 0|1
+
+Set -handle_trailing_comment => 1 to enable support of parameter trailing
+comments.
+
+For example, if we have a parameter line like this:
+
+ param1=value1;comment1
+
+by default, handle_trailing_comment will be set to B<0>, and we will get
+I<value1;comment1> as the value of I<param1>. If we have 
+-handle_trailing_comment set to B<1>, then we will get I<value1> 
+as the value for I<param1>, and I<comment1> as the trailing comment of 
+I<param1>. 
+
+Set and get methods for trailing comments are provided as
+L</SetParameterTrailingComment> and L</GetParameterTrailingComment>.
+
 =back
 
 =cut
@@ -293,10 +355,14 @@ sub new {
 
   my $self = bless {
 	default => '',
+	fallback =>undef,
+	fallback_used => 0,
 	imported =>undef,
 	v =>{},
 	cf => undef,
 	firstload => 1,
+	nomultiline => 0,
+    handle_trailing_comment => 0,
   }, $class;
 
   if( ref($parms{-import}) && ($parms{-import}->isa('Config::IniFiles')) ) {
@@ -329,17 +395,23 @@ sub new {
   if (defined ($v = delete $parms{'-default'})) {
     $self->{default} = $self->{nocase} ? lc($v) : $v;
   }
+  if (defined ($v = delete $parms{'-fallback'})) {
+    $self->{fallback} = $self->{nocase} ? lc($v) : $v;
+  }
   if (defined ($v = delete $parms{'-reloadwarn'})) {
     $self->{reloadwarn} = $v ? 1 : 0;
   }
+  if (defined ($v = delete $parms{'-nomultiline'})) {
+    $self->{nomultiline} = $v ? 1 : 0;
+  }
   if (defined ($v = delete $parms{'-allowcontinue'})) {
     $self->{allowcontinue} = $v ? 1 : 0;
   }
   if (defined ($v = delete $parms{'-allowempty'})) {
-	  $self->{allowempty} = $v ? 1 : 0;
+     $self->{allowempty} = $v ? 1 : 0;
   }
   if (defined ($v = delete $parms{'-negativedeltas'})) {
-	  $self->{negativedeltas} = $v ? 1 : 0;
+      $self->{negativedeltas} = $v ? 1 : 0;
   }
   if (defined ($v = delete $parms{'-commentchar'})) {
     if(!defined $v || length($v) != 1) {
@@ -365,6 +437,11 @@ sub new {
       $self->{allowed_comment_char} = $v;
     }
   }
+
+  if (defined ($v = delete $parms{'-handle_trailing_comment'})) {
+      $self->{handle_trailing_comment} = $v ? 1 : 0;
+  }
+
   $self->{comment_char} = '#' unless exists $self->{comment_char};
   $self->{allowed_comment_char} = ';' unless exists $self->{allowed_comment_char};
   # make sure that comment character is always allowed
@@ -653,12 +730,25 @@ sub _nextline {
 	if (!exists $self->{line_ends}) {
 		# no $self->{line_ends} is a hint set by caller that we are at
 		# the first line (kludge kludge).
-		do {
-			local $/=\1; my $nextchar=<$fh>;
-			return undef if (!defined $nextchar);
-			$_ .= $nextchar;
-		} until (m/(\015\012?|\012|\025|\n)$/s);
-		$self->{line_ends}=$1;
+		{
+			local $/=\1;
+			my $nextchar;
+			do {
+				$nextchar=<$fh>;
+				return undef if (!defined $nextchar);
+				$_ .= $nextchar;
+			} until (m/((\015|\012|\025|\n)$)/s);
+			$self->{line_ends}=$1;
+			if ($nextchar eq "\x0d") {
+				# peek at the next char
+				$nextchar = <$fh>;
+				if ($nextchar eq "\x0a") {
+					$self->{line_ends} .= "\x0a";
+				} else {
+					seek $fh, -1, 1;
+				}
+			}
+		}
 
 		# If there's a UTF BOM (Byte-Order-Mark) in the first
 		# character of the first line then remove it before processing
@@ -694,6 +784,7 @@ sub ReadConfig {
   my($group, $groupmem);
   my($parm, $val);
   my @cmts;
+  my $end_comment;
 
   @Config::IniFiles::errors = ( );
 
@@ -707,6 +798,7 @@ sub ReadConfig {
   $self->{EOT}    = {};
   $self->{mysects} = []; # A pair of hashes to remember which params are loaded
   $self->{myparms} = {}; # or set using the API vs. imported - useful for
+  $self->{peCMT}  = {};  # this will store trailing comments at the end of single-lined params
   # import shadowing, see below, and WriteConfig(-delta=>1)
 
   if( defined $self->{imported} ) {
@@ -728,6 +820,7 @@ sub ReadConfig {
   );
   
   my $nocase = $self->{nocase};
+  my $end_commenthandle = $self->{handle_trailing_comment};
 
   # If this is a reload and we want warnings then send one to the STDERR log
   unless( $self->{firstload} || !$self->{reloadwarn} ) {
@@ -748,7 +841,8 @@ sub ReadConfig {
   }
   
   # Get mod time of file so we can retain it (if not from STDIN)
-  if (ref($fh) ne "IO::Scalar")
+  # also check if it's a real file (could have been a filehandle made from a scalar).
+  if (ref($fh) ne "IO::Scalar" && -e $fh)
   {
     my @stats = stat $fh;
     $self->{file_mode} = sprintf("%04o", $stats[2]) if defined $stats[2];
@@ -791,6 +885,11 @@ sub ReadConfig {
       @cmts = ();
     }
     elsif (($parm, $val) = /^\s*([^=]*?[^=\s])\s*=\s*(.*)$/) {	# new parameter
+		if ((!defined($sect)) and defined($self->{fallback}))
+		{ 
+			$sect = $self->{fallback};
+			$self->{fallback_used}++;
+		}
 		if (!defined $sect) {
 			CORE::push(@Config::IniFiles::errors, sprintf('%d: %s', $lineno,
 				qq#parameter found outside a section#));
@@ -832,6 +931,16 @@ sub ReadConfig {
 	  $lineno++;
           $val .= $_;
         }
+
+        # we should split value and comments if there is any comment
+        if ($end_commenthandle &&
+            $val =~ /(.*?)\s*[$allCmt]\s*([^$allCmt\s]*)$/) {
+            $val = $1;
+            $end_comment = $2;
+        } else {
+            $end_comment = "";
+        }
+
 		@val = $val;
       }
 		# Now load value
@@ -847,6 +956,8 @@ sub ReadConfig {
 		$self->SetParameterComment($sect, $parm, @cmts);
 		@cmts = ( );
 		$self->SetParameterEOT($sect,$parm,$eotmark) if (defined $eotmark);
+        # if handle_trailing_comment is off, this line makes no sense, since all $end_comment=""
+        $self->SetParameterTrailingComment($sect, $parm, $end_comment);
 
     } else {
       CORE::push(@Config::IniFiles::errors, sprintf("Line \%d in file " . $self->{cf} . " is mal-formed:\n\t\%s", $lineno, $_));
@@ -1313,6 +1424,56 @@ should be set to 1 if writing only delta.
 
 =cut
 
+sub _OutputParam {
+    my ($self, $sect, $parm, $val, $ors, $end_comment, $output_cb) = @_;
+
+    if (! @$val) {
+        # An empty variable - see:
+        # https://rt.cpan.org/Public/Bug/Display.html?id=68554
+        $output_cb->("$parm=");
+        $output_cb->("$ors");
+    }
+    elsif ((@$val == 1) or $self->{nomultiline}) {
+        my $cnt = 0;
+        foreach (@{$val}) {
+            $cnt++;
+            $output_cb->("$parm=$_");
+            # output trailing comment at the last parameter
+            if ($end_comment && $cnt == @$val) {
+                $output_cb->(" $self->{comment_char} $end_comment") 
+            }
+            $output_cb->("$ors");
+        }
+    }
+    else
+    {
+        my $eotmark = $self->{EOT}{$sect}{$parm} || 'EOT';
+
+        # Make sure the $eotmark does not occur inside the string.
+        my @letters = ('A' .. 'Z');
+        my $joined_val = join(q{ }, @$val);
+        while (index($joined_val, $eotmark) >= 0)
+        {
+            $eotmark .= $letters[rand(@letters)];
+        }
+
+        $output_cb->("$parm= <<$eotmark$ors");
+        my $cnt = 0;
+        foreach (@{$val}) {
+            $cnt++;
+            $output_cb->("$_");
+            # output trailing comment at the last parameter
+            if ($end_comment && $cnt == @$val) {
+                $output_cb->(" $self->{comment_char} $end_comment")
+            }
+            $output_cb->("$ors");
+        }
+        $output_cb->("$eotmark$ors");
+    }
+
+    return;
+}
+
 sub OutputConfig {
     my ($self, $delta) = @_;
 
@@ -1339,7 +1500,13 @@ sub OutputConfig {
                 print "$_$ors";
             }
         }
-        print "[$sect]$ors";
+
+        if (!
+            ($self->{fallback_used} and $sect eq $self->{fallback})
+        )
+        {
+            print "[$sect]$ors";
+        }
         next unless ref $self->{v}{$sect} eq 'HASH';
 
         PARM:
@@ -1360,51 +1527,34 @@ sub OutputConfig {
             }
 
             my $val = $self->{v}{$sect}{$parm};
+            my $end_comment = $self->{peCMT}{$sect}{$parm};
+
             next if ! defined ($val); # No parameter exists !!
-            if (ref($val) eq 'ARRAY') {
-                my $eotmark = $self->{EOT}{$sect}{$parm} || 'EOT';
-                print "$parm= <<$eotmark$ors";
-                foreach (@{$val}) {
-                    print "$_$ors";
-                }
-                print "$eotmark$ors";
-            } elsif( $val =~ /[$ors]/ ) {
-                # The FETCH of a tied hash is never called in 
-                # an array context, so generate a EOT multiline
-                # entry if the entry looks to be multiline
-                my @val = split /[$ors]/, $val, -1;
-                if( @val > 1 ) {
-                    my $eotmark = $self->{EOT}{$sect}{$parm} || 'EOT';
 
-                    # Make sure the $eotmark does not occur inside the string.
-                    my @letters = ('A' .. 'Z');
-                    while (index($val, $eotmark) >= 0)
-                    {
-                        $eotmark .= $letters[rand(@letters)];
-                    }
-
-                    print "$parm= <<$eotmark$ors";
-                    print map "$_$ors", @val;
-                    print "$eotmark$ors";
-                } else {
-                    print "$parm=$val[0]$ors";
-                } # end if
-            } else {
-                print "$parm=$val$ors";
-            }
+            $self->_OutputParam(
+                $sect,
+                $parm, 
+                ((ref($val) eq 'ARRAY')
+                    ? $val
+                    : [split /[$ors]/, $val, -1]
+                ),
+                $ors,
+                defined $end_comment ? $end_comment : "",
+                sub { print @_; },
+            );
         }
     }
     foreach my $comment ($self->_GetEndComments()) {
         print "$comment$ors";
     }
     return 1;
-}
+    }
 
 =head2 SetSectionComment($section, @comment)
 
 Sets the comment for section $section to the lines contained in @comment.
 
-Each comment line will be prepended with the comment charcter (default
+Each comment line will be prepended with the comment character (default
 is C<#>) if it doesn't already have a comment character (ie: if the
 line does not start with whitespace followed by an allowed comment
 character, default is C<#> and C<;>).
@@ -1712,6 +1862,68 @@ sub DeleteParameterEOT
 	delete $self->{EOT}{$sect}{$parm};
 }
 
+=head2 SetParameterTrailingComment ($section, $parameter, $cmt)
+
+Set the end trailing comment for the given section and parameter. 
+If there is a old comment for the parameter, it will be 
+overwritten by the new one.
+
+If there is a new parameter trailing comment to be added, the
+value should be added first.
+
+=cut
+
+sub SetParameterTrailingComment
+{
+    my $self = shift;
+    my $sect = shift;
+    my $parm = shift;
+    my $cmt = shift;
+
+    return undef if not defined $sect;
+    return undef if not defined $parm;
+    return undef if not defined $cmt;
+
+    if ($self->{nocase}) {
+        $sect = lc($sect);
+        $parm = lc($parm);
+    }
+
+    # confirm the parameter exist
+    return undef if not exists $self->{v}{$sect}{$parm};
+
+    $self->_touch_parameter($sect, $parm);
+    $self->{peCMT}{$sect}{$parm} = $cmt;
+    1;
+}
+
+=head2 GetParameterTrailingComment ($section, $parameter)
+
+An accessor method to read the trailing comment after the parameter. 
+The trailing comment will be returned if there is one. A null string 
+will be returned if the parameter exists but no comment for it. 
+otherwise, L<undef> will be returned.
+
+=cut
+
+sub GetParameterTrailingComment
+{
+    my $self = shift;
+    my $sect = shift;
+    my $parm = shift;
+
+    return undef if not defined $sect;
+    return undef if not defined $parm;
+
+    if ($self->{nocase}) {
+        $sect = lc($sect);
+        $parm = lc($parm);
+    }
+
+    # confirm the parameter exist
+    return undef if not exists $self->{v}{$sect}{$parm};
+    return $self->{peCMT}{$sect}{$parm};
+}
 
 =head2 Delete
 
@@ -2052,7 +2264,7 @@ sub _make_filehandle {
 
   if (ref($thing) eq "SCALAR") {
 	  if (eval { require IO::Scalar; $IO::Scalar::VERSION >= 2.109; }) {
-		  return new IO::Scalar($thing);
+		  return IO::Scalar->new($thing);
 	  } else {
 		  warn "SCALAR reference as file descriptor requires IO::stringy ".
 			"v2.109 or later" if ($^W);
@@ -2289,8 +2501,8 @@ those coming from the ``original'' one and the lines coming from the
 file, the latter taking precedence. For example, let's say that
 C<$master> and C<overlay> were created like this:
 
-   my $master  = new Config::IniFiles(-file => "master.ini");
-   my $overlay = new Config::IniFiles(-file => "overlay.ini",
+   my $master  = Config::IniFiles->new(-file => "master.ini");
+   my $overlay = Config::IniFiles->new(-file => "overlay.ini",
 			-import => $master);
 
 If the contents of C<master.ini> and C<overlay.ini> are respectively
@@ -2318,9 +2530,9 @@ file for a Perl application, that can be overridden piecewise by a
 much shorter, per-site configuration file. Assuming UNIX-style path
 names, this would be done like this:
 
-   my $defaultconfig=new Config::IniFiles
+   my $defaultconfig = Config::IniFiles->new
        (-file => "/usr/share/myapp/myapp.ini.default");
-   my $config=new Config::IniFiles
+   my $config = Config::IniFiles->new
        (-file => "/etc/myapp.ini", -import => $defaultconfig);
    # Now use $config and forget about $defaultconfig in the rest of
    # the program
@@ -2346,7 +2558,7 @@ C<overlay.ini> would now contain
 This is called a I<delta file> (see L</WriteConfig>). The untouched
 [section2] and arg0 do not appear, and the config file is therefore
 shorter; while of course, reloading the configuration into C<$master>
-and C<$overlay>, either through C<$overlay->ReadConfig()> or through
+and C<$overlay>, either through C<< $overlay->ReadConfig() >> or through
 the same code as above (e.g. when application restarts), would yield
 exactly the same result had the overlay object been saved in whole to
 the file system.
